import os
import sys
import json
import webbrowser
import asyncio
from dotenv import load_dotenv
from .finic import Finic
import subprocess
import argparse
from .copilot import copilot

load_dotenv()

def get_api_key() -> str:
    
    # Check if API key exists
    env_path = os.path.join(os.getcwd(), '.env')
    if os.path.exists(env_path):
        load_dotenv(env_path)
    api_key = os.getenv('FINIC_API_KEY')
    
    if not api_key:
        # Open browser for login
        print("No Finic API key found. Create a new account? (y/n)")
        create_account = input()
        if create_account == "y":
            print("Please log in to Finic in the opened browser.")
            print("After logging in, copy the API key from the Settings page and paste it here.")
            webbrowser.open('https://app.finic.io/settings')
        
        print("Enter your API key:")
        api_key = input()
        
        # Save API key to .env file
        with open('.env', 'a') as env_file:
            env_file.write(f'\nFINIC_API_KEY={api_key}')
        
        print("API key has been saved to .env file.")
    
    return api_key

def zip_files_cli(zip_file):
    try:
        # First zip command (for untracked files)
        command_1 = (
            f"git ls-files -z --others --exclude-standard | xargs -0 zip -r {zip_file}"
        )
        subprocess.run(
            command_1, shell=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
        )

        # Second zip command (for tracked files)
        command_2 = f"git ls-files -z | xargs -0 zip -ur {zip_file}"
        subprocess.run(
            command_2, shell=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
        )

        print(f"Zipped files into {zip_file}")

    except subprocess.CalledProcessError as e:
        print(f"Error occurred during zipping: {e}")


def finic_init():
    current_directory = os.getcwd()
    new_directory_path = os.path.join(current_directory, "finic_tasks")
    
    # Check if the directory already exists
    if os.path.exists(new_directory_path):
        print("Error: The directory 'finic_tasks' already exists.")
        return
    
    # Create the new directory
    os.makedirs(new_directory_path)
    
    # Change to the new directory
    os.chdir(new_directory_path)

    os.system(
        f"git clone --depth 1 https://github.com/finic-ai/finic-tasks-boilerplate . && rm -rf .git"
    )
    print(
        f"The /finic_tasks directory has been created.\ncd into it and run `poetry install` to install dependencies, then `finic copilot` to create your first task."
    )


def deploy():
    api_key = get_api_key()

    # Check if finic_config.json exists
    if not os.path.exists("finic_config.json"):
        print(
            "finic_config.json not found. Please create a finic_config.json file in the root directory of your project"
        )
        return

    server_url = None
    with open("finic_config.json", "r") as f:
        config = json.load(f)
        if "finic_url" in config:
            server_url = config["finic_url"]
        if "agent_name" not in config:
            print("Please specify the agent_name in the finic_config.json file")
            return
        agent_name = config["agent_name"]
        if "agent_id" not in config:
            print("Please specify the agent_id in the finic_config.json file")
            return
        agent_id = config["agent_id"]
        if "num_retries" not in config:
            print("Please specify the num_retries in the finic_config.json file")
            return
        num_retries = config["num_retries"]

    finic = Finic(api_key=api_key, finic_url=server_url)

    temp_dir = os.path.join(os.getcwd(), "temp")
    zip_file = os.path.join(temp_dir, "project.zip")

    # Create the /tmp directory if it doesn't exist
    os.makedirs(temp_dir, exist_ok=True)

    # Zip the project into /tmp/project.zip, ignoring all .gitignore patterns

    zip_files_cli(zip_file)

    result = finic.deploy_agent(agent_id, agent_name, num_retries, zip_file)

    if result:
        print(f"Agent with ID {agent_id} deployed successfully")
    else:
        print("Agent deployment failed")

def main():
    parser = argparse.ArgumentParser(description="CLI for Finic's python library.")
    subparsers = parser.add_subparsers(dest='command', required=True)

    # Deploy command
    deploy_parser = subparsers.add_parser(
        'deploy', 
        help='Deploy the agent to Finic cloud'
    )

    # Connect to remote browser command
    connect_to_browser_parser = subparsers.add_parser(
        'connect', 
        help='Connects to a remote browser'
    )

    connect_to_browser_parser.add_argument(
        '--cdp-url', 
        help='The URL of the browser to connect to',
        required=True
    )
    connect_to_browser_parser.add_argument(
        '--api-key', 
        help='A Finic API key',
        required=True
    )

    # Initialize Finic directory command
    init_parser = subparsers.add_parser(
        'init', 
        help='Initialize a Finic directory'
    )
    

    # Generate selectors command
    generate_parser = subparsers.add_parser(
        'copilot', 
        help='Start the Finic copilot'
    )
    generate_parser.add_argument(
        '--url', 
        help='The URL of the starting page of the workflow',
        required=True
    )

    args = parser.parse_args()

    finic_api_key = get_api_key()

    if args.command == 'deploy':
        deploy()
    elif args.command == 'init':
        finic_init()
    elif args.command == 'copilot':
        asyncio.run(copilot(args.url, finic_api_key))

if __name__ == "__main__":
    main()


from playwright.async_api import async_playwright, Playwright, Page, ElementHandle, CDPSession
import asyncio
from enum import Enum
from typing import List, Dict, Any, Optional
from bs4 import BeautifulSoup, Tag
import shutil
from pydantic import BaseModel, field_validator
import uuid
import json
import base64
from typing import Literal, Tuple
import requests
import platform
import os
import sys
import itertools

class NodeDetails(BaseModel):
    selector: Optional[str] = None
    nodeId: int
    backendNodeId: int
    nodeType: int
    nodeName: str
    localName: str
    nodeValue: str
    childNodeCount: Optional[int] = 0
    attributes: List[Dict[str, str]] = []
    textContent: Optional[str] = None
    outerHTML: Optional[str] = None

    @field_validator('attributes', mode='before')
    def parse_attributes(cls, v):
        if isinstance(v, list):
            if all(isinstance(item, str) for item in v):
                return [{'name': v[i], 'value': v[i+1]} for i in range(0, len(v), 2) if i+1 < len(v)]
            elif all(isinstance(item, dict) for item in v):
                return v
        raise ValueError('attributes must be a list of strings or dictionaries')

def test_selector(selector: str, page: Page) -> int:
    elements = page.query_selector_all(selector)
    return len(elements)

def generate_selector(node: NodeDetails, page: Page, cdp_session: CDPSession, dom_snapshot: List[Dict[str, Any]]) -> str:
    xpath_root = f"//{node.localName}"
    
    node_tree = dom_snapshot["documents"][0]["nodes"]
    selected_node_index = node_tree["backendNodeId"].index(node.backendNodeId)
    parent_index = node_tree["parentIndex"][selected_node_index]

    valid_xpaths = []
    
    def generate_by_attribute(node: NodeDetails, max_attributes: int = 8) -> List[str]:
        xpaths_to_test = []
        if node.attributes:
            # Create a list of xpaths using combinations of attributes
            for i in range(1, min(max_attributes, len(node.attributes) + 1)):
                for combo in itertools.combinations(node.attributes, i):
                    attr_conditions = []
                    for attr in combo:
                        attr_conditions.append(f"@{attr['name']}='{attr['value']}'")
                    xpath = f"{xpath_root}[{' and '.join(attr_conditions)}]"
                    xpaths_to_test.append(xpath)
        
        return xpaths_to_test
    
    def generate_by_classnames(node: NodeDetails, max_classnames: int = 8) -> List[str]:
        xpaths_to_test = []
        if node.attributes:
            class_attributes = [attr for attr in node.attributes if attr["name"] == "class"]
            if class_attributes:
                classnames = class_attributes[0]["value"].split()
                xpaths_to_test = []
                for i in range(1, min(max_classnames, len(classnames) + 1)):
                    for combo in itertools.combinations(classnames, i):
                        class_condition = " and ".join([f"contains(@class, '{c}')" for c in combo])
                        xpaths_to_test.append(f"{xpath_root}[{class_condition}]")
        return xpaths_to_test
    
    def generate_by_text_content(node: NodeDetails) -> List[str]:
        xpaths_to_test = []
        if node.textContent:
            xpaths_to_test.append(f'{xpath_root}[normalize-space(string())="{node.textContent}"]')
        return xpaths_to_test
    
    async def generate_by_sibling(node: NodeDetails) -> List[str]:
        xpaths_to_test = []
        if node.textContent:
            xpaths_to_test.append(f'{xpath_root}/following-sibling::*[normalize-space(string())="{node.textContent}"]')
        return xpaths_to_test

    async def generate_by_path(node: NodeDetails) -> List[str]:
        # Recursively go up the tree until a unique identifier is found, or until we find the <body> tag
        xpaths_to_test = []
        node_index = node_tree["backendNodeId"].index(node.backendNodeId)
        parent_node_backend_id = node_tree["parentIndex"][node_index]
        parent_node_details = await cdp_session.send("DOM.describeNode", {"backendNodeId": parent_node_backend_id})
        parent_node = NodeDetails(**parent_node_details["node"])
        
        outer_html = await cdp_session.send("DOM.getOuterHTML", {"backendNodeId": parent_node_backend_id})
        parent_node.outerHTML = outer_html["outerHTML"]
        parent_node.textContent = BeautifulSoup(parent_node.outerHTML, 'html.parser').get_text(separator=' ', strip=True)

        return xpaths_to_test

    try:
        xpaths_to_test = (generate_by_attribute(node))
        for xpath in xpaths_to_test:
            if test_selector(xpath, page) == 1:
                valid_xpaths.append(xpath)
        
        xpaths_to_test = (generate_by_classnames(node))
        for xpath in xpaths_to_test:
            if test_selector(xpath, page) == 1:
                valid_xpaths.append(xpath)

        xpaths_to_test = (generate_by_text_content(node))
        for xpath in xpaths_to_test:
            if test_selector(xpath, page) == 1:
                valid_xpaths.append(xpath)

        xpaths_to_test = (generate_by_path(node))
        for xpath in xpaths_to_test:
            if test_selector(xpath, page) == 1:
                valid_xpaths.append(xpath)

    except Exception as e:
        print(e)

async def enable_inspection(cdp_session: CDPSession):
    await cdp_session.send('Overlay.setInspectMode', {'mode': 'searchForNode', 'highlightConfig': {'showInfo': True, 'showExtensionLines': True, 'contentColor': {'r': 255, 'g': 81, 'b': 6, 'a': 0.2}}})

async def disable_inspection(cdp_session: CDPSession):
    await cdp_session.send('Overlay.setInspectMode', {'mode': 'none', 'highlightConfig': {}})

def print_welcome_message():
    # Clear the terminal
    print("\033[2J\033[H", end="")
    print("\033[38;2;255;81;6m" + """
    ███████╗  ██╗  ███╗   ██╗  ██╗  ██████╗
    ██╔════╝  ██║  ████╗  ██║  ██║  ██╔════╝
    █████╗    ██║  ██╔██╗ ██║  ██║  ██║     
    ██╔══╝    ██║  ██║╚██╗██║  ██║  ██║     
    ██║       ██║  ██║ ╚████║  ██║  ╚██████╗
    ╚═╝       ╚═╝  ╚═╝  ╚═══╝  ╚═╝   ╚═════╝
    """ + "\033[0m")
    print("\033[1m\033[38;2;255;165;0mGenerate code to automate any website using AI.\033[0m\n")
    print("Type 'help' for a list of commands. View full instructions here: https://docs.finic.ai/capture-mode")

async def handle_process_node(finic_api_key: str, task_name: str, intent: str, selected_node: List[NodeDetails], cdp_session: CDPSession, page: Page) -> None:
    # Generate a selector for the selected node
    selector = generate_selector(selected_node[0])
    selected_node[0].selector = selector
    generated_code = None

    task_file_path = f"finic_tasks/{task_name}.py"
    with open(task_file_path, 'r') as f:
        existing_code = f.read()

    # Make a GET request to the Finic server to get the code
    try:
        headers = {
            "Authorization": f"Bearer {finic_api_key}"  # Assuming api_key is available in this scope
        }
        body = {
            "intent": intent,
            "element": selected_node[0],
            "existing_code": existing_code
        }
        response = requests.post("https://api.finic.ai/copilot", headers=headers, json=body)
        response.raise_for_status()  # Raise an exception for bad status codes
        generated_code = response.json()["code"]
        # You can process the copilot_data here as needed
    except requests.RequestException as e:
        print(f"Error making request to Finic AI Copilot API: {e}")
    
    # Write the generated code to the appropriate file
    with open(task_file_path, 'w') as f:
        f.write(existing_code + generated_code)

    
    

def print_element_to_terminal(element: NodeDetails):
    terminal_width, _ = shutil.get_terminal_size()
    box_width = min(80, terminal_width - 2)  # Max 80, or 2 less than terminal width
    text_content = element.textContent[:box_width-4] if element.textContent else ""  # Truncate to fit
    attributes = ' '.join([f"{attribute['name']}=\"{attribute['value']}\"" 
                           for attribute in element.attributes])

    # Print the box
    print("\033[1m\033[94m┌─" + "─" * (box_width - 2) + "┐\033[0m")
    print(f"\033[1m\033[94m│\033[0m {'Currently selected element':^{box_width-2}}\033[1m\033[94m│\033[0m")
    print("\033[1m\033[94m├─" + "─" * (box_width - 2) + "┤\033[0m")
    print(f"\033[1m\033[94m│\033[0m \033[1m\033[92mTag:\033[0m {element.localName:<{box_width-7}}\033[1m\033[94m│\033[0m")
    print(f"\033[1m\033[94m│\033[0m \033[1m\033[92mID:\033[0m {element.backendNodeId:<{box_width-6}}\033[1m\033[94m│\033[0m")
    print(f"\033[1m\033[94m│\033[0m \033[1m\033[92mAttributes:\033[0m{' ':{box_width-14}}\033[1m\033[94m│\033[0m")
    
    # Print attributes, wrapping if necessary
    attr_lines = [attributes[i:i+box_width-4] for i in range(0, len(attributes), box_width-4)]
    for line in attr_lines:
        print(f"\033[1m\033[94m│\033[0m {line:<{box_width-2}}\033[1m\033[94m│\033[0m")
    
    print(f"\033[1m\033[94m│\033[0m \033[1m\033[92mText content:\033[0m{' ':{box_width-15}}\033[1m\033[94m│\033[0m")
    print(f"\033[1m\033[94m│\033[0m {text_content:<{box_width-2}}\033[1m\033[94m│\033[0m")
    print("\033[1m\033[94m└─" + "─" * (box_width - 2) + "┘\033[0m")

async def cycle_elements_up(cdp_session: CDPSession, selected_node: List[NodeDetails], dom_snapshot: List[Dict[str, Any]]) -> None:
    node_tree = dom_snapshot["documents"][0]["nodes"]
    selected_node_index = node_tree["backendNodeId"].index(selected_node[0].backendNodeId)
    parent_index = node_tree["parentIndex"][selected_node_index]

    if parent_index:
        parent_backend_id = node_tree["backendNodeId"][parent_index]
        await handle_inspect_node(cdp_session, selected_node, backend_node_id=parent_backend_id)

async def cycle_elements_down(cdp_session: CDPSession, selected_node: List[NodeDetails], dom_snapshot: List[Dict[str, Any]]) -> None:
    node_tree = dom_snapshot["documents"][0]["nodes"]
    selected_node_index = node_tree["backendNodeId"].index(selected_node[0].backendNodeId)
    try:
        child_index = node_tree["parentIndex"].index(selected_node_index)
        child_backend_id = node_tree["backendNodeId"][child_index]
        await handle_inspect_node(cdp_session, selected_node, backend_node_id=child_backend_id)
    except ValueError:
        return

async def handle_inspect_node(cdp_session: CDPSession, selected_node: List[NodeDetails], node_id: Optional[int] = None, backend_node_id: Optional[int] = None):
    if not node_id and not backend_node_id:
        raise ValueError("Either node_id or backend_node_id must be provided")
    # Request the document so DOM.pushNodesByBackendIdsToFrontend works
    document = await cdp_session.send("DOM.getDocument")

    if backend_node_id and not node_id:
        # Find the node in the DOM
        await cdp_session.send("DOM.pushNodesByBackendIdsToFrontend", {"backendNodeIds": [backend_node_id]})
        node_details = await cdp_session.send("DOM.describeNode", {"backendNodeId": backend_node_id})
        node_id = node_details["node"]["nodeId"]
    outer_html = await cdp_session.send("DOM.getOuterHTML", {"nodeId": node_id})

    await cdp_session.send("DOM.setInspectedNode", {"nodeId": node_id})
    await cdp_session.send("DOM.highlightNode", {
        "highlightConfig": {
            "showInfo": True,
            "showExtensionLines": True,
            "containerColor": {"r": 0, "g": 255, "b": 0, "a": 0.3},
            "contentColor": {"r": 0, "g": 255, "b": 0, "a": 0.3}
        },
        "nodeId": node_id
    })
    result = NodeDetails(**node_details["node"])
    result.outerHTML = outer_html["outerHTML"]
    result.textContent = BeautifulSoup(outer_html["outerHTML"], 'html.parser').get_text(separator=' ', strip=True)

    selected_node[0] = result
    print_welcome_message()
    print_element_to_terminal(result)
    print("Describe the action to be taken on this element: ", end="", flush=True)

def create_artifacts(task_name: str) -> Tuple[str, str]:
    # Check if the finic_tasks directory exists in the current directory
    current_directory = os.getcwd()
    finic_tasks_path = os.path.join(current_directory, 'finic_tasks')
    finic_config_path = os.path.join(current_directory, 'finic_config.yaml')
    
    if os.path.exists(finic_tasks_path) and os.path.exists(finic_config_path):
        task_folder_path = os.path.join(finic_tasks_path, task_name)
        os.makedirs(task_folder_path, exist_ok=True)

        task_file_path = os.path.join(task_folder_path, f'{task_name}.py')
        selectors_path = os.path.join(task_folder_path, 'selectors.yaml')
        
        if not os.path.exists(selectors_path):
            with open(selectors_path, 'w') as f:
                f.write("")

        if not os.path.exists(task_file_path):
            with open(task_file_path, 'w') as f:
                f.write(f"""from finic import Finic
from playwright.sync_api import Page

def main(page: Page, finic: Finic):
""")

        return task_file_path, selectors_path
    else:
        print("This is not a Finic project. Run `finic init` to initialize Finic.")
        sys.exit(0)

async def copilot(url: str, finic_api_key: str):
    print_welcome_message()
    task_name = input("\n\nGive your task a unique name (e.g. 'automate_tax_website'): ")
    # Replace spaces and dashes with underscores in task_name
    task_name = task_name.replace(' ', '_').replace('-', '_')
    task_file_path, selectors_path = create_artifacts(task_name)
    print(f"Created task file: {task_file_path}")

    inspection_mode = True
    selected_node: List[NodeDetails] = [None]
    dom_snapshot = None

    playwright = await async_playwright().start()
    
    browser = await playwright.chromium.launch(headless=False, devtools=True)
    page = await browser.new_page()

    ### SET UP CDP AND EVENT LISTENERS ###
    cdp_session = await page.context.new_cdp_session(page)
    await cdp_session.send('DOM.enable')
    await cdp_session.send('Overlay.enable')
    await cdp_session.send('Runtime.enable')
    await cdp_session.send('DOMSnapshot.enable')


    await cdp_session.send('Overlay.setInspectMode', {
        'mode': 'searchForNode',
        'highlightConfig': {'showInfo': True, 'showExtensionLines': True, 'contentColor': {'r': 255, 'g': 81, 'b': 6, 'a': 0.2}}
    })

    async def handle_key_event(event):
        nonlocal inspection_mode
        modifier = 5  # Represents Ctrl+Shift
        # Toggle inspect mode
        if event.get('type') == 'keyDown' and event.get('code') == 'KeyF' and event.get('modifiers') == modifier:
            if inspection_mode:
                await disable_inspection(cdp_session)
                inspection_mode = False
                print("Switched to interaction mode. Use CMD/CTRL+SHIFT+F to re-enable selection mode.")
            else:
                await enable_inspection(cdp_session)
                inspection_mode = True
                print("Switched to selection mode. Use CMD/CTRL+SHIFT+F to re-enable interaction mode.")
        
        # Cycle through element layers
        elif len(selected_node) > 0 and inspection_mode:
            if event.get('type') == 'keyDown' and event.get('code') == 'ArrowUp':
                await cycle_elements_up(cdp_session, selected_node, dom_snapshot)
            elif event.get('type') == 'keyDown' and event.get('code') == 'ArrowDown':
                await cycle_elements_down(cdp_session, selected_node, dom_snapshot)

    def handle_console_event(event: Dict[str, Any]):
        if event['type'] == 'log':
            data = event.get('args', [])
            if data and data[0].get('value', '').startswith('KeyEvent:'):
                key_event = json.loads(data[1]['value'])
                asyncio.create_task(handle_key_event(key_event))

    cdp_session.on("Overlay.inspectNodeRequested", lambda event: handle_inspect_node(cdp_session, selected_node, backend_node_id=event["backendNodeId"]))
    cdp_session.on("Runtime.consoleAPICalled", lambda params: handle_console_event(params))

    # Navigate to a website
    await page.goto(url)
    await page.wait_for_load_state("load")

    await page.evaluate("""
        window.addEventListener('keydown', (event) => {
            const key = event.key;
            const code = event.code;
            const modifiers = (event.metaKey ? 8 : 0) | (event.ctrlKey ? 4 : 0) | (event.altKey ? 2 : 0) | (event.shiftKey ? 1 : 0);
            console.log('KeyEvent:', JSON.stringify({ type: 'keyDown', key, code, modifiers }));
            
            if (code === 'ArrowUp' || code === 'ArrowDown') {
                event.preventDefault();
            }
        });
    """)
    print("\nSelect an element in the browser. Enter 'mode' to toggle between selection and interaction mode, or 'quit' to exit: ", end="", flush=True)
    
    dom_snapshot = await cdp_session.send("DOMSnapshot.captureSnapshot", {"computedStyles": []})

    while True:
        user_input = await asyncio.get_event_loop().run_in_executor(None, input)
        if user_input.lower() in ['quit', 'q']:
            browser.close()
            break
        elif user_input.lower() in ['mode', 'm']:
            if inspection_mode:
                await cdp_session.send('Overlay.setInspectMode', {'mode': 'none', 'highlightConfig': {}})
                inspection_mode = False
                print("Switched to interaction mode.")
            else:
                await cdp_session.send('Overlay.setInspectMode', {'mode': 'searchForNode', 'highlightConfig': {'showInfo': True, 'showExtensionLines': True, 'contentColor': {'r': 255, 'g': 81, 'b': 6, 'a': 0.2}}})
                inspection_mode = True
                print("Switched to selection mode.")
        elif user_input.lower() in ['help', 'h']:
            print("Click on eleements in the browser. Confirm your selection in the box above. Use 'a|add' to queue an element for generation. Use 'g|generate' to generate selectors for queued elements. Use 'l|list' to view queued elements.")
            print("\n\033[1mCommands:\033[0m")
            print("  • \033[1m'm'|'mode'\033[0m    - Change between interaction and selection mode")
            print("  • \033[1m'quit'|'q'\033[0m    - Quit the program")
            print("\n\033[1m\033[38;2;255;165;0mEnter command:\033[0m ", end="", flush=True)
        else:
            if len(selected_node) > 0:
                pass
                # result = await handle_process_node(finic_api_key, task_name, user_input, selected_node, cdp_session, page)
            else:
                print("Invalid input. Please select an element in the browser first or enter a command.")
    
    browser.close()



from typing import Dict, Optional, List
from functools import wraps
import json
import yaml
import os
from enum import Enum
import requests
import datetime
import sys
from pydantic import BaseModel
from typing import Any, Union, Literal
from playwright.sync_api import sync_playwright, Browser, BrowserContext, Page, ElementHandle
from playwright.async_api import async_playwright, Browser as AsyncBrowser, BrowserContext as AsyncBrowserContext, Page as AsyncPage, ElementHandle as AsyncElementHandle
from dotenv import load_dotenv
from typing import Tuple


class FinicEnvironment(str, Enum):
    LOCAL = "local"
    PROD = "production"
    DEV = "development"


class LogSeverity(str, Enum):
    DEFAULT = "DEFAULT"
    WARNING = "WARNING"
    ERROR = "ERROR"


class ExecutionLog(BaseModel):
    severity: LogSeverity
    message: str
    timestamp: Optional[datetime.datetime] = None

    class Config:
        json_encoders = {datetime: lambda v: v.isoformat() if v else None}


class ExecutionAttempt(BaseModel):
    success: bool
    attempt_number: int
    logs: List[ExecutionLog] = []


class LogExecutionAttemptRequest(BaseModel):
    execution_id: str
    agent_id: str
    results: Dict[str, Any]
    attempt: ExecutionAttempt

class AuthException(Exception):
    pass

class SelectorException(Exception):
    pass

class StdoutLogger:
    def __init__(self, original_stdout):
        self.logs: List[ExecutionLog] = []
        self.original_stdout = original_stdout

    def write(self, message):
        if message.strip():  # Avoid logging empty messages
            log = ExecutionLog(
                severity=LogSeverity.DEFAULT,
                timestamp=datetime.datetime.now(datetime.timezone.utc),
                message=message.strip(),
            )
            self.logs.append(log)
            self.original_stdout.write(
                f"{log.timestamp} [{log.severity.value}] {log.message}\n"
            )

    def flush(self):
        pass

    def get_logs(self):
        return self.logs

class SelectorService:
    def __init__(self, selectors: Dict[str, str]):
        self.selectors = selectors

    def get(self, key: str) -> str:
        return self.selectors.get(key)
    
    def set(self, key: str, value: str):
        self.selectors[key] = value


CONTEXT_STORAGE_PATH = "browser_state.json"

class Finic:
    api_key: Optional[str] = None
    env: Optional[FinicEnvironment] = None
    finic_url: Optional[str] = None
    selectors: Optional[SelectorService] = None

    def __init__(
        self,
        api_key: Optional[str] = None,
        environment: Optional[FinicEnvironment] = None,
        finic_url: Optional[str] = None,
        selector_source: Literal["file", "cloud"] = None,
        selector_file_path: Optional[str] = "selectors.yaml"
    ):

        default_env = os.environ.get("FINIC_ENVIRONMENT") or FinicEnvironment.LOCAL
        self.environment = environment if environment else FinicEnvironment(default_env)
        if selector_source == "file":
            with open(selector_file_path, 'r') as file:
                self.selectors = SelectorService(yaml.safe_load(file))
        elif selector_source == "cloud":
            raise NotImplementedError("Cloud selector source not implemented")
        if api_key:
            self.api_key = api_key
        else:
            self.api_key = os.getenv("FINIC_API_KEY")
        self.finic_url = finic_url or "https://api.finic.io"

    def get_agent_input(self) -> Optional[Dict[str, Any]]:
        if self.environment == FinicEnvironment.LOCAL:
            path = os.path.join(os.getcwd(), "finic_input.json")
            if os.path.exists(path):
                with open(path, 'r') as f:
                    return json.load(f)
            else:
                return None
        else:
            load_dotenv()
            if os.getenv("FINIC_INPUT"):
                return json.loads(os.getenv("FINIC_INPUT"))
            return None
    
    def save_browser_context(self, context: BrowserContext, browser_id: Optional[str] = None):
        if self.environment == FinicEnvironment.LOCAL:
            path = os.path.join(os.getcwd(), CONTEXT_STORAGE_PATH)
            context.storage_state(path=path)
            print(f"Browser context saved to: {path}")
        else:
            browser_id = browser_id or os.getenv("FINIC_BROWSER_ID")
            if not browser_id:
                return
            browser_context = context.storage_state()
            response = requests.post(
                f"{self.finic_url}/browser-state/{browser_id}",
                headers={"Authorization": f"Bearer {self.api_key}"},
                json=browser_context
            )
            response.raise_for_status()
            browser = response.json()
            print(f"Browser context saved for browser: {browser['id']}")

    def get_browser_context(self, browser_id: Optional[str] = None) -> Optional[Dict[str, Any]]:
        if self.environment == FinicEnvironment.LOCAL:
            path = os.path.join(os.getcwd(), CONTEXT_STORAGE_PATH)
            if os.path.exists(path):
                with open(path, 'r') as f:
                    return json.load(f)
            else:
                return None
        else:
            if not browser_id:
                browser_id = os.getenv("FINIC_BROWSER_ID")
            response = requests.get(
                f"{self.finic_url}/browser-state/{browser_id}",
                headers={"Authorization": f"Bearer {self.api_key}"},
            )
            response.raise_for_status()
            browser = response.json()
            return browser.get('state')
        
    def save_session_results(self, results: Dict):
        if self.environment == FinicEnvironment.LOCAL:
            path = os.path.join(os.getcwd(), f"results.json")
            with open(path, 'w') as f:
                json.dump(results, f, ensure_ascii=False, indent=4, default=lambda o: o.dict())
            print(f"Session results saved to: {path}")
        else:
            session_id = os.getenv("FINIC_SESSION_ID")
            response = requests.post(
                f"{self.finic_url}/session/{session_id}",
                headers={"Authorization": f"Bearer {self.api_key}"},
                json={"results": results}
            )
            response.raise_for_status()
            print(f"Session results saved for session: {session_id}")

    def get_session_results(self, session_id: Optional[str] = None) -> Optional[List[Dict]]:
        if self.environment == FinicEnvironment.LOCAL:
            path = os.path.join(os.getcwd(), f"results.json")
            if os.path.exists(path):    
                with open(path, 'r') as f:
                    return json.load(f)
            else:
                return None
        else:
            if not session_id:
                session_id = os.getenv("FINIC_SESSION_ID")
            response = requests.get(
                f"{self.finic_url}/session/{session_id}",
                headers={"Authorization": f"Bearer {self.api_key}"},
            )
            response.raise_for_status()
            response_json = response.json()
            if response_json:
                return response_json.get("results", [])
            return []
    
    def launch_browser_sync(self, **kwargs) -> Tuple[Page, BrowserContext]:
        video_dir = os.path.join(os.getcwd(), "session_recording")
        # If it doesn't exist, create it
        if not os.path.exists(video_dir):
            os.makedirs(video_dir)
        else:
            # Delete all files in the video directory
            for file in os.listdir(video_dir):
                os.remove(os.path.join(video_dir, file))
        playwright = sync_playwright().start()
        browser = playwright.chromium.launch(**kwargs)
        saved_context = self.get_browser_context()
        context = browser.new_context(storage_state=saved_context, record_video_dir=video_dir)
        context.add_init_script("""
            Object.defineProperty(navigator, 'webdriver', {
                get: () => undefined
            });
        """)

        
        initial_page = context.new_page()
        recording_path_file = os.getenv("FINIC_SESSION_RECORDING_PATH")
        if recording_path_file:
            # Save the video path to the file
            with open(recording_path_file, "w") as f:
                path = initial_page.video.path()
                f.write(str(path))

        def disable_video(page: Page):
            page.on("close", lambda p: p.video.delete())

        context.on("page", disable_video)

        return initial_page, context
    
    async def launch_browser_async(self,  **kwargs) -> AsyncBrowserContext:
        video_dir = os.path.join(os.getcwd(), "session_recording")
        # If it doesn't exist, create it
        if not os.path.exists(video_dir):
            os.makedirs(video_dir)
        else:
            # Delete all files in the video directory
            for file in os.listdir(video_dir):
                os.remove(os.path.join(video_dir, file))
        playwright = await async_playwright().start()
        browser = await playwright.chromium.launch(**kwargs)
        saved_context = self.get_browser_context()
        context = await browser.new_context(storage_state=saved_context, record_video_dir=video_dir)
        context.add_init_script("""
            Object.defineProperty(navigator, 'webdriver', {
                get: () => undefined
            });
        """)

        
        initial_page = await context.new_page()
        recording_path_file = os.getenv("FINIC_SESSION_RECORDING_PATH")
        if recording_path_file:
            # Save the video path to the file
            with open(recording_path_file, "w") as f:
                path = initial_page.video.path()
                f.write(str(path))

        def disable_video(page: Page):
            page.on("close", lambda p: p.video.delete())

        context.on("page", disable_video)

        return initial_page, context
    
    @classmethod
    def screenshot(self, page: Union[Page, AsyncPage], selector: str, path: str) -> None:
        # Create a new div for screenshot
        page.evaluate("""() => {
            const outerContainer = document.createElement('div');
            outerContainer.id = 'screenshotOuterContainer';
            outerContainer.style.position = 'fixed';
            outerContainer.style.zIndex = '2147483647'; // Maximum z-index value
            outerContainer.style.top = '0';
            outerContainer.style.left = '0';
            outerContainer.style.width = '100%';
            outerContainer.style.height = '100%';
            outerContainer.style.backgroundColor = 'white';
            outerContainer.style.display = 'flex';
            outerContainer.style.justifyContent = 'center';
            outerContainer.style.alignItems = 'center';

            const innerContainer = document.createElement('div');
            innerContainer.id = 'screenshotInnerContainer';
            innerContainer.style.display = 'inline-block';

            outerContainer.appendChild(innerContainer);
            document.body.appendChild(outerContainer);
        }""")

        # Move the ad into the screenshot container
        cloned_element = page.evaluate_handle("""(selector) => {
            const innerContainer = document.getElementById('screenshotInnerContainer');
            const adElement = document.evaluate(selector, document, null, XPathResult.FIRST_ORDERED_NODE_TYPE, null).singleNodeValue;
            
            if (innerContainer && adElement) {
                const clone = adElement.cloneNode(true);
                const styles = window.getComputedStyle(adElement);
                
                clone.style.cssText = Array.from(styles).reduce((str, property) => {
                    return `${str}${property}:${styles.getPropertyValue(property)};`;
                }, '');
                
                // Reset positioning styles
                clone.style.position = 'static';
                clone.style.margin = '0';
                clone.style.width = 'auto';
                clone.style.height = 'auto';
                
                innerContainer.appendChild(clone);
                return clone;
            }
            return null;
        }""", selector)

        page.wait_for_selector(selector)  # Wait for the element to be added
        cloned_element.screenshot(path=path)

        # Remove the screenshot container div
        page.evaluate("""() => {
            const outerContainer = document.getElementById('screenshotOuterContainer');
            outerContainer.remove();
        }""")

    @classmethod
    def entrypoint(cls, func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            config = {}
            if os.path.exists('finic_config.json'):
                with open('finic_config.json', 'r') as config_file:
                    config = json.load(config_file)
            inputs = config.get("args", {})

            result = func(*args, **kwargs, **inputs)

            with open('output.json', 'w') as output_file:
                json.dump(result, output_file)

            return result
        return wrapper

    @classmethod
    def procedure(cls, func):
        @wraps(func)
        def wrapper(page, *args, **kwargs):
            if not isinstance(page, Page):
                raise TypeError("The first argument must be a Page object")
            result = func(page, *args, **kwargs)
            if not isinstance(result, Page):
                raise TypeError("The function must return a Page object")
            return result
        return wrapper

    @classmethod
    def max_scroll(self, element: Union[Page, ElementHandle], timeout: int = 2000):
        last_height = element.evaluate('document.body.scrollHeight')
        while True:
            element.evaluate('window.scrollTo(0, document.body.scrollHeight)')
            element.wait_for_timeout(timeout)
            new_height = element.evaluate('document.body.scrollHeight')
            if new_height == last_height:
                break
            last_height = new_height
        return new_height
    
    def deploy_agent(self, agent_id: str, agent_name: str, num_retries: int, zip_file: str) -> bool:
        with open(zip_file, 'rb') as f:
            upload_file = f.read()
        response = requests.post(
            f"{self.finic_url}/agent-upload-link/{agent_id}",
            headers={"Authorization": f"Bearer {self.api_key}"},
            json={"agent_id": agent_id, "agent_name": agent_name, "num_retries": num_retries}
        )
        response.raise_for_status()
        upload_url = response.json()["upload_url"]
        response = requests.put(upload_url, data=upload_file)
        response.raise_for_status()
        return True
    
    
